package bou.amine.apps.readerforselfossv2.android

import android.content.Intent
import android.graphics.Color
import android.graphics.drawable.Drawable
import android.graphics.drawable.GradientDrawable
import android.net.Uri
import android.os.Bundle
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.ImageView
import android.widget.Toast
import androidx.appcompat.app.ActionBarDrawerToggle
import androidx.appcompat.app.AlertDialog
import androidx.appcompat.app.AppCompatActivity
import androidx.appcompat.widget.SearchView
import androidx.core.view.doOnNextLayout
import androidx.drawerlayout.widget.DrawerLayout
import androidx.recyclerview.widget.*
import androidx.work.Constraints
import androidx.work.ExistingPeriodicWorkPolicy
import androidx.work.PeriodicWorkRequestBuilder
import androidx.work.WorkManager
import bou.amine.apps.readerforselfossv2.android.adapters.ItemCardAdapter
import bou.amine.apps.readerforselfossv2.android.adapters.ItemListAdapter
import bou.amine.apps.readerforselfossv2.android.adapters.ItemsAdapter
import bou.amine.apps.readerforselfossv2.android.background.LoadingWorker
import bou.amine.apps.readerforselfossv2.android.databinding.ActivityHomeBinding
import bou.amine.apps.readerforselfossv2.android.settings.SettingsActivity
import bou.amine.apps.readerforselfossv2.android.themes.AppColors
import bou.amine.apps.readerforselfossv2.android.themes.Toppings
import bou.amine.apps.readerforselfossv2.android.utils.bottombar.maybeShow
import bou.amine.apps.readerforselfossv2.android.utils.bottombar.removeBadge
import bou.amine.apps.readerforselfossv2.android.utils.customtabs.CustomTabActivityHelper
import bou.amine.apps.readerforselfossv2.repository.Repository
import bou.amine.apps.readerforselfossv2.model.SelfossModel
import bou.amine.apps.readerforselfossv2.service.AppSettingsService
import bou.amine.apps.readerforselfossv2.utils.*
import com.ashokvarma.bottomnavigation.BottomNavigationBar
import com.ashokvarma.bottomnavigation.BottomNavigationItem
import com.ashokvarma.bottomnavigation.TextBadgeItem
import com.bumptech.glide.Glide
import com.bumptech.glide.request.RequestOptions
import com.ftinc.scoop.Scoop
import com.mikepenz.aboutlibraries.LibsBuilder
import com.mikepenz.materialdrawer.holder.BadgeStyle
import com.mikepenz.materialdrawer.holder.ColorHolder
import com.mikepenz.materialdrawer.holder.StringHolder
import com.mikepenz.materialdrawer.model.DividerDrawerItem
import com.mikepenz.materialdrawer.model.PrimaryDrawerItem
import com.mikepenz.materialdrawer.model.ProfileDrawerItem
import com.mikepenz.materialdrawer.model.SecondaryDrawerItem
import com.mikepenz.materialdrawer.model.interfaces.*
import com.mikepenz.materialdrawer.util.AbstractDrawerImageLoader
import com.mikepenz.materialdrawer.util.DrawerImageLoader
import com.mikepenz.materialdrawer.util.addStickyFooterItem
import com.mikepenz.materialdrawer.util.updateBadge
import com.mikepenz.materialdrawer.widget.AccountHeaderView
import com.russhwolf.settings.Settings
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import org.kodein.di.DIAware
import org.kodein.di.android.closestDI
import org.kodein.di.instance
import java.util.concurrent.TimeUnit
import kotlin.concurrent.thread

class HomeActivity : AppCompatActivity(), SearchView.OnQueryTextListener, DIAware {

    private val DRAWER_ID_TAGS = 100101L
    private val DRAWER_ID_HIDDEN_TAGS = 101100L
    private val DRAWER_ID_SOURCES = 100110L
    private val DRAWER_ID_FILTERS = 100111L

    private var items: ArrayList<SelfossModel.Item> = ArrayList()

    private var elementsShown: ItemType = ItemType.UNREAD
    private var lastFetchDone: Boolean = false

    private lateinit var tabNewBadge: TextBadgeItem
    private lateinit var tabArchiveBadge: TextBadgeItem
    private lateinit var tabStarredBadge: TextBadgeItem
    private lateinit var customTabActivityHelper: CustomTabActivityHelper
    private lateinit var appColors: AppColors
    private var offset: Int = 0
    private var firstVisible: Int = 0
    private lateinit var recyclerViewScrollListener: RecyclerView.OnScrollListener
    private lateinit var binding: ActivityHomeBinding

    private var recyclerAdapter: RecyclerView.Adapter<*>? = null

    private var fromTabShortcut: Boolean = false

    private lateinit var tagsBadge: Map<Long, Int>

    override val di by closestDI()
    private val repository : Repository by instance()
    private val appSettingsService : AppSettingsService by instance()

    data class DrawerData(val tags: List<SelfossModel.Tag>?, val sources: List<SelfossModel.Source>?)

    override fun onStart() {
        super.onStart()
        customTabActivityHelper.bindCustomTabsService(this)
    }

    override fun onCreate(savedInstanceState: Bundle?) {

        // Add appcolors to DI
        appColors = AppColors(this@HomeActivity)

        super.onCreate(savedInstanceState)
        binding = ActivityHomeBinding.inflate(layoutInflater)
        val view = binding.root

        fromTabShortcut =  intent.getIntExtra("shortcutTab", -1) != -1
        repository.offlineOverride =  intent.getBooleanExtra("startOffline", false)

        if (fromTabShortcut) {
            elementsShown = ItemType.fromInt(intent.getIntExtra("shortcutTab", ItemType.UNREAD.position))
        }

        setContentView(view)

        handleThemeBinding()

        setSupportActionBar(binding.toolBar)
        supportActionBar?.setDisplayHomeAsUpEnabled(true)
        supportActionBar?.setHomeButtonEnabled(true)
        val mDrawerToggle = ActionBarDrawerToggle(this, binding.drawerContainer, binding.toolBar, R.string.material_drawer_open, R.string.material_drawer_close)
        binding.drawerContainer.addDrawerListener(mDrawerToggle)
        mDrawerToggle.syncState()

        customTabActivityHelper = CustomTabActivityHelper()

        handleBottomBar()
        handleDrawer()

        handleSwipeRefreshLayout()

        getElementsAccordingToTab()


        CoroutineScope(Dispatchers.Main).launch {
            repository.tryToCacheItemsAndGetNewOnes()
        }

    }

    private fun handleSwipeRefreshLayout() {
        binding.swipeRefreshLayout.setColorSchemeResources(
            R.color.refresh_progress_1,
            R.color.refresh_progress_2,
            R.color.refresh_progress_3
        )
        binding.swipeRefreshLayout.setOnRefreshListener {
            repository.offlineOverride = false
            lastFetchDone = false
            handleDrawerItems()
            CoroutineScope(Dispatchers.Main).launch {
                getElementsAccordingToTab()
                binding.swipeRefreshLayout.isRefreshing = false
            }
        }

        val simpleItemTouchCallback =
            object : ItemTouchHelper.SimpleCallback(
                0,
                ItemTouchHelper.LEFT or ItemTouchHelper.RIGHT
            ) {
                override fun getSwipeDirs(
                    recyclerView: RecyclerView,
                    viewHolder: RecyclerView.ViewHolder
                ): Int =
                    if (elementsShown == ItemType.STARRED) {
                        0
                    } else {
                        super.getSwipeDirs(
                            recyclerView,
                            viewHolder
                        )
                    }

                override fun onMove(
                    recyclerView: RecyclerView,
                    viewHolder: RecyclerView.ViewHolder,
                    target: RecyclerView.ViewHolder
                ): Boolean = false

                override fun onSwiped(viewHolder: RecyclerView.ViewHolder, swipeDir: Int) {
                    val position = viewHolder.bindingAdapterPosition
                    val i = items.elementAtOrNull(position)

                    if (i != null) {
                        val adapter = binding.recyclerView.adapter as ItemsAdapter<*>

                        adapter.handleItemAtIndex(position)

                        reloadBadgeContent()

                        val tagHashes = i.tags.map { it.longHash() }
                        tagsBadge = tagsBadge.map {
                            if (tagHashes.contains(it.key)) {
                                (it.key to (it.value - 1))
                            } else {
                                (it.key to it.value)
                            }
                        }.toMap()
                        reloadTagsBadges()

                        // Just load everythin
                        if (items.size <= 0) {
                            getElementsAccordingToTab()
                        }
                    } else {
                        Toast.makeText(
                            this@HomeActivity,
                            "Found null when swiping at positon $position.",
                            Toast.LENGTH_LONG
                        ).show()
                    }
                }
            }

        ItemTouchHelper(simpleItemTouchCallback).attachToRecyclerView(binding.recyclerView)
    }

    private fun handleBottomBar() {

        tabNewBadge = TextBadgeItem()
            .setText("")
            .setHideOnSelect(false).hide(false)
            .setBackgroundColor(appColors.colorPrimary)
        tabArchiveBadge = TextBadgeItem()
            .setText("")
            .setHideOnSelect(false).hide(false)
            .setBackgroundColor(appColors.colorPrimary)
        tabStarredBadge = TextBadgeItem()
            .setText("")
            .setHideOnSelect(false).hide(false)
            .setBackgroundColor(appColors.colorPrimary)

        val tabNew =
            BottomNavigationItem(
                R.drawable.ic_tab_fiber_new_black_24dp,
                getString(R.string.tab_new)
            ).setActiveColor(appColors.colorAccent)
                .setBadgeItem(tabNewBadge)
        val tabArchive =
            BottomNavigationItem(
                R.drawable.ic_tab_archive_black_24dp,
                getString(R.string.tab_read)
            ).setActiveColor(appColors.colorAccentDark)
                .setBadgeItem(tabArchiveBadge)
        val tabStarred =
            BottomNavigationItem(
                R.drawable.ic_tab_favorite_black_24dp,
                getString(R.string.tab_favs)
            ).setActiveColorResource(R.color.pink)
                .setBadgeItem(tabStarredBadge)

        binding.bottomBar
            .addItem(tabNew)
            .addItem(tabArchive)
            .addItem(tabStarred)
            .setFirstSelectedPosition(0)
            .initialise()
        binding.bottomBar.setMode(BottomNavigationBar.MODE_SHIFTING)
        binding.bottomBar.setBackgroundStyle(BottomNavigationBar.BACKGROUND_STYLE_STATIC)

        if (fromTabShortcut) {
            binding.bottomBar.selectTab(elementsShown.position - 1)
        }
    }

    override fun onResume() {
        super.onResume()

        appColors = AppColors(this@HomeActivity)

        handleDrawerItems()

        handleThemeUpdate()

        reloadLayoutManager()

        if (appSettingsService.isInfiniteLoadingEnabled()) {
            handleInfiniteScroll()
        } else {
            binding.recyclerView.setHasFixedSize(true)
        }

        handleBottomBarActions()

        handleRecurringTask()

        CoroutineScope(Dispatchers.Main).launch {
            repository.handleDBActions()
        }

        getElementsAccordingToTab()
    }

    override fun onStop() {
        super.onStop()
        customTabActivityHelper.unbindCustomTabsService(this)
    }

    private fun handleThemeBinding() {
        val scoop = Scoop.getInstance()
        scoop.bind(this, Toppings.PRIMARY.value, binding.toolBar)
        scoop.bindStatusBar(this, Toppings.PRIMARY_DARK.value)
    }

    private fun handleThemeUpdate() {

        val scoop = Scoop.getInstance()
        scoop.update(Toppings.PRIMARY.value, appColors.colorPrimary)

        scoop.update(Toppings.PRIMARY_DARK.value, appColors.colorPrimaryDark)
    }

    private fun handleDrawer() {
        DrawerImageLoader.init(object : AbstractDrawerImageLoader() {
            override fun set(imageView: ImageView, uri: Uri, placeholder: Drawable, tag: String?) {
                Glide.with(this@HomeActivity)
                        .asBitmap()
                        .load(uri)
                        .apply(RequestOptions()
                                .placeholder(R.mipmap.ic_launcher)
                                .fallback(R.mipmap.ic_launcher)
                                .fitCenter())
                        .into(imageView)
            }

            override fun cancel(imageView: ImageView) {
                Glide.with(this@HomeActivity).clear(imageView)
            }
        })

        val drawerListener = object : DrawerLayout.DrawerListener {
            override fun onDrawerSlide(drawerView: View, slideOffset: Float) {
                // We do nothing
            }

            override fun onDrawerOpened(drawerView: View) {
                binding.bottomBar.hide()
            }

            override fun onDrawerClosed(drawerView: View) {
                binding.bottomBar.show()
            }

            override fun onDrawerStateChanged(newState: Int) {
                // We do nothing
            }

        }

        binding.drawerContainer.addDrawerListener(drawerListener)

        binding.mainDrawer.addStickyFooterItem(
            PrimaryDrawerItem().apply {
                nameRes = R.string.drawer_report_bug
                iconRes = R.drawable.ic_bug_report_black_24dp
                isIconTinted = true
                onDrawerItemClickListener = { _, _, _ ->
                    val browserIntent = Intent(Intent.ACTION_VIEW, Uri.parse(AppSettingsService.trackerUrl))
                    startActivity(browserIntent)
                    false
                }
            })

        binding.mainDrawer.addStickyFooterItem(
            PrimaryDrawerItem().apply {
                nameRes = R.string.title_activity_settings
                iconRes = R.drawable.ic_settings_black_24dp
                isIconTinted = true
                onDrawerItemClickListener = { _, _, _ ->
                    startActivity(Intent(this@HomeActivity, SettingsActivity::class.java))
                    false
                }
            })

        if (appSettingsService.isDisplayAccountHeaderEnabled()) {
            AccountHeaderView(this).apply {
                attachToSliderView(binding.mainDrawer)
                addProfiles(
                    ProfileDrawerItem().apply {
                        nameText = appSettingsService.getBaseUrl()
                        setBackgroundResource(R.drawable.bg)
                        iconRes = R.mipmap.ic_launcher
                        selectionListEnabledForSingleProfile = false
                    }
                )
            }
        }
    }

    // TODO: refactor this.
    // TODO: use updateSources
    private fun handleDrawerItems() {
        tagsBadge = emptyMap()
        fun handleDrawerData(maybeDrawerData: DrawerData?, loadedFromCache: Boolean = false) {
            fun createDrawerItem(
                it: SelfossModel.Tag
            ) {
                val gd = GradientDrawable()
                val gdColor = try {
                    Color.parseColor(it.color)
                } catch (e: IllegalArgumentException) {
                    appColors.colorPrimary
                }

                gd.setColor(gdColor)
                gd.shape = GradientDrawable.RECTANGLE
                gd.setSize(30, 30)
                gd.cornerRadius = 30F

                val drawerItem = PrimaryDrawerItem()
                    .apply {
                        nameText = it.tag.getHtmlDecoded()
                        identifier = it.tag.longHash()
                        iconDrawable = gd
                        badgeStyle = BadgeStyle().apply {
                            textColor = ColorHolder.fromColor(Color.WHITE)
                            color = ColorHolder.fromColor(appColors.colorAccent)
                        }
                        onDrawerItemClickListener = { _, _, _ ->
                            repository.tagFilter = it
                            repository.sourceFilter = null
                            getElementsAccordingToTab()
                            fetchOnEmptyList()
                            false
                        }
                    }
                if (it.unread > 0) {
                    drawerItem.badgeText = it.unread.toString()
                }

                binding.mainDrawer.itemAdapter.add(drawerItem)
            }

            fun handleTags(maybeTags: List<SelfossModel.Tag>?) {
                if (maybeTags == null) {
                    if (loadedFromCache) {
                        binding.mainDrawer.itemAdapter.add(
                            SecondaryDrawerItem()
                                .apply { nameRes = R.string.drawer_error_loading_tags; isSelectable = false }
                        )
                    }
                } else {
                    val filteredTags = maybeTags
                        .filterNot { appSettingsService.getHiddenTags().contains(it.tag) }
                        .sortedBy { it.unread == 0 }
                    tagsBadge = filteredTags.map {
                        createDrawerItem(it)

                        (it.tag.longHash() to it.unread)
                    }.toMap()
                }
            }

            fun handleHiddenTags(maybeTags: List<SelfossModel.Tag>?) {
                if (maybeTags == null) {
                    if (loadedFromCache) {
                        binding.mainDrawer.itemAdapter.add(
                            SecondaryDrawerItem().apply {
                                nameRes = R.string.drawer_error_loading_tags
                                isSelectable = false
                            }
                        )
                    }
                } else {
                    val filteredHiddenTags: List<SelfossModel.Tag> =
                        maybeTags.filter { appSettingsService.getHiddenTags().contains(it.tag) }
                    tagsBadge = filteredHiddenTags.map {
                        createDrawerItem(it)

                        (it.tag.longHash() to it.unread)
                    }.toMap()
                }
            }

            fun handleSources(maybeSources: List<SelfossModel.Source>?) {
                if (maybeSources == null) {
                    if (loadedFromCache) {
                        binding.mainDrawer.itemAdapter.add(
                            SecondaryDrawerItem().apply {
                                nameRes = R.string.drawer_error_loading_sources
                                isSelectable = false
                            }
                        )
                    }
                } else {
                    for (source in maybeSources) {
                        val item = PrimaryDrawerItem().apply {
                            nameText = source.title.getHtmlDecoded()
                            identifier = source.id.toLong()
                            iconUrl = source.getIcon(repository.baseUrl)
                            onDrawerItemClickListener = { _,_,_ ->
                                repository.sourceFilter = source
                                repository.tagFilter = null
                                getElementsAccordingToTab()
                                fetchOnEmptyList()
                                false
                            }
                        }
                        binding.mainDrawer.itemAdapter.add(item)
                    }
                }
            }

            binding.mainDrawer.itemAdapter.clear()
            if (maybeDrawerData != null) {
                binding.mainDrawer.itemAdapter.add(
                    SecondaryDrawerItem().apply {
                        nameRes = R.string.drawer_item_filters
                        isSelectable = false
                        identifier = DRAWER_ID_FILTERS
                        badgeRes = R.string.drawer_action_clear
                        onDrawerItemClickListener = { _,_,_ ->
                            repository.sourceFilter = null
                            repository.tagFilter = null
                            binding.mainDrawer.setSelectionAtPosition(-1)
                            getElementsAccordingToTab()
                            fetchOnEmptyList()
                            false
                        }
                    }
                )
                if (appSettingsService.getHiddenTags().isNotEmpty()) {
                    binding.mainDrawer.itemAdapter.add(
                        DividerDrawerItem(),
                        SecondaryDrawerItem().apply {
                            nameRes = R.string.drawer_item_hidden_tags
                            identifier = DRAWER_ID_HIDDEN_TAGS
                            isSelectable = false
                        }
                    )
                    handleHiddenTags(maybeDrawerData.tags)
                }
                binding.mainDrawer.itemAdapter.add(
                    DividerDrawerItem(),
                    SecondaryDrawerItem().apply {
                        nameRes = R.string.drawer_item_tags
                        identifier = DRAWER_ID_TAGS
                        isSelectable = false
                    }
                )
                handleTags(maybeDrawerData.tags)
                binding.mainDrawer.itemAdapter.add(
                    DividerDrawerItem(),
                    SecondaryDrawerItem().apply {
                        nameRes = R.string.drawer_item_sources
                        identifier = DRAWER_ID_SOURCES
                        isSelectable = false
                        badgeRes = R.string.drawer_action_edit
                        onDrawerItemClickListener = { v,_,_ ->
                            startActivity(Intent(v!!.context, SourcesActivity::class.java))
                            false
                        }
                    }
                )
                handleSources(maybeDrawerData.sources)
                binding.mainDrawer.itemAdapter.add(
                    DividerDrawerItem(),
                    PrimaryDrawerItem().apply {
                        nameRes = R.string.action_about
                        isSelectable = false
                        iconRes = R.drawable.ic_info_outline_white_24dp
                        isIconTinted = true
                        onDrawerItemClickListener = { _,_,_ ->
                            LibsBuilder()
                                .withAboutIconShown(true)
                                .withAboutVersionShown(true)
                                .start(this@HomeActivity)
                            false
                        }
                    }
                )

                if (!loadedFromCache) {
                    if (maybeDrawerData.tags != null) {
                        thread {
                            repository.resetDBTagsWithData(maybeDrawerData.tags)
                        }
                    }
                    if (maybeDrawerData.sources != null) {
                        thread {
                            repository.resetDBSourcesWithData(maybeDrawerData.sources)
                        }
                    }
                }
            } else {
                if (!loadedFromCache) {
                    binding.mainDrawer.itemAdapter.add(
                        PrimaryDrawerItem().apply {
                            nameRes = R.string.no_tags_loaded
                            identifier = DRAWER_ID_TAGS
                            isSelectable = false
                        },
                        PrimaryDrawerItem().apply {
                            nameRes = R.string.no_sources_loaded
                            identifier = DRAWER_ID_SOURCES
                            isSelectable = false
                        }
                    )
                }
            }
        }

        fun drawerApiCalls(maybeDrawerData: DrawerData?) {
            var tags: List<SelfossModel.Tag>? = null
            var sources: List<SelfossModel.Source>?

            fun sourcesApiCall() {
                CoroutineScope(Dispatchers.Main).launch {
                    val response = repository.getSources()
                    if (response != null) {
                        sources = response
                        val apiDrawerData = DrawerData(tags, sources)
                        if ((maybeDrawerData != null && maybeDrawerData != apiDrawerData) || maybeDrawerData == null) {
                            handleDrawerData(apiDrawerData)
                        }
                    } else {
                        val apiDrawerData = DrawerData(tags, null)
                        if ((maybeDrawerData != null && maybeDrawerData != apiDrawerData) || maybeDrawerData == null) {
                            handleDrawerData(apiDrawerData)
                        }
                    }
                }
            }

            CoroutineScope(Dispatchers.IO).launch {
                tags = repository.getTags()
                sourcesApiCall()
            }
        }

        binding.mainDrawer.itemAdapter.add(
            PrimaryDrawerItem().apply {
                nameRes = R.string.drawer_loading
                isSelectable = false
            }
        )

        thread {
            val drawerData = DrawerData(repository.getDBTags().map { it.toView() },
                                        repository.getDBSources().map { it.toView() })
            runOnUiThread {
                handleDrawerData(drawerData, loadedFromCache = true)
                drawerApiCalls(drawerData)
            }
        }
    }

    private fun reloadLayoutManager() {
        val currentManager = binding.recyclerView.layoutManager
        val layoutManager: RecyclerView.LayoutManager

        // This will only update the layout manager if settings changed
        when (currentManager) {
            is StaggeredGridLayoutManager ->
                if (!appSettingsService.isCardViewEnabled()) {
                    layoutManager = GridLayoutManager(
                        this,
                        calculateNoOfColumns()
                    )
                    binding.recyclerView.layoutManager = layoutManager
                }
            is GridLayoutManager ->
                if (appSettingsService.isCardViewEnabled()) {
                    layoutManager = StaggeredGridLayoutManager(
                        calculateNoOfColumns(),
                        StaggeredGridLayoutManager.VERTICAL
                    )
                    layoutManager.gapStrategy =
                            StaggeredGridLayoutManager.GAP_HANDLING_MOVE_ITEMS_BETWEEN_SPANS
                    binding.recyclerView.layoutManager = layoutManager
                }
            else ->
                if (currentManager == null) {
                    if (!appSettingsService.isCardViewEnabled()) {
                        layoutManager = GridLayoutManager(
                            this,
                            calculateNoOfColumns()
                        )
                        binding.recyclerView.layoutManager = layoutManager
                    } else {
                        layoutManager = StaggeredGridLayoutManager(
                            calculateNoOfColumns(),
                            StaggeredGridLayoutManager.VERTICAL
                        )
                        layoutManager.gapStrategy =
                                StaggeredGridLayoutManager.GAP_HANDLING_MOVE_ITEMS_BETWEEN_SPANS
                        binding.recyclerView.layoutManager = layoutManager
                    }
                }
        }
    }

    private fun handleBottomBarActions() {
        binding.bottomBar.setTabSelectedListener(object : BottomNavigationBar.OnTabSelectedListener {
            override fun onTabUnselected(position: Int) = Unit

            override fun onTabReselected(position: Int) {

                when (val layoutManager = binding.recyclerView.adapter) {
                    is StaggeredGridLayoutManager ->
                        if (layoutManager.findFirstCompletelyVisibleItemPositions(null)[0] == 0) {
                            getElementsAccordingToTab()
                        } else {
                            layoutManager.scrollToPositionWithOffset(0, 0)
                        }
                    is GridLayoutManager ->
                        if (layoutManager.findFirstCompletelyVisibleItemPosition() == 0) {
                            getElementsAccordingToTab()
                        } else {
                            layoutManager.scrollToPositionWithOffset(0, 0)
                        }
                    else -> Unit
                }
            }

            override fun onTabSelected(position: Int) {
                offset = 0
                lastFetchDone = false

                elementsShown = ItemType.fromInt(position + 1)
                getElementsAccordingToTab()
                binding.recyclerView.scrollToPosition(0)

                fetchOnEmptyList()
            }
        })
    }

    private fun fetchOnEmptyList() {
        binding.recyclerView.doOnNextLayout {
            // TODO: do if last element (or is empty ?)
            getElementsAccordingToTab(true)
        }
    }

    private fun handleInfiniteScroll() {
        recyclerViewScrollListener = object : RecyclerView.OnScrollListener() {
            override fun onScrolled(localRecycler: RecyclerView, dx: Int, dy: Int) {
                if (dy > 0) {
                    val lastVisibleItem = getLastVisibleItem()

                    if (lastVisibleItem == (items.size - 1) && items.size < maxItemNumber()) {
                        getElementsAccordingToTab(appendResults = true)
                    }
                }
            }
        }

        binding.recyclerView.clearOnScrollListeners()
        binding.recyclerView.addOnScrollListener(recyclerViewScrollListener)
    }

    private fun getLastVisibleItem() : Int {
        return when (val manager = binding.recyclerView.layoutManager) {
            is StaggeredGridLayoutManager -> manager.findLastCompletelyVisibleItemPositions(
                null
            ).last()
            is GridLayoutManager -> manager.findLastCompletelyVisibleItemPosition()
            else -> 0
        }
    }

    private fun mayBeEmpty() =
        if (items.isEmpty()) {
            binding.emptyText.visibility = View.VISIBLE
        } else {
            binding.emptyText.visibility = View.GONE
        }

    private fun getElementsAccordingToTab(
        appendResults: Boolean = false
    ) {
        offset = if (appendResults && items.size > 0) {
            items.size - 1
        } else {
            0
        }
        firstVisible = if (appendResults) firstVisible else 0

        getItems(appendResults, elementsShown)
    }

    private fun getItems(appendResults: Boolean, itemType: ItemType) {
        CoroutineScope(Dispatchers.Main).launch {
            binding.swipeRefreshLayout.isRefreshing = true
            repository.displayedItems = itemType
            items = if (appendResults) {
                repository.getOlderItems()
            } else {
                repository.getNewerItems()
            }
            binding.swipeRefreshLayout.isRefreshing = false
            handleListResult()
        }
    }

    private fun handleListResult(appendResults: Boolean = false) {
        if (appendResults) {
            val oldManager = binding.recyclerView.layoutManager
            firstVisible = when (oldManager) {
                is StaggeredGridLayoutManager ->
                    oldManager.findFirstCompletelyVisibleItemPositions(null).last()
                is GridLayoutManager ->
                    oldManager.findFirstCompletelyVisibleItemPosition()
                else -> 0
            }
        }

        if (recyclerAdapter == null) {
            if (appSettingsService.isCardViewEnabled()) {
                recyclerAdapter =
                        ItemCardAdapter(
                            this,
                            items,
                            customTabActivityHelper,
                            appColors,
                        ) {
                            updateItems(it)
                        }
            } else {
                recyclerAdapter =
                        ItemListAdapter(
                            this,
                            items,
                            customTabActivityHelper,
                            appColors,
                        ) {
                            updateItems(it)
                        }

                binding.recyclerView.addItemDecoration(
                    DividerItemDecoration(
                        this@HomeActivity,
                        DividerItemDecoration.VERTICAL
                    )
                )
            }
            binding.recyclerView.adapter = recyclerAdapter
        } else {
                (recyclerAdapter as ItemsAdapter<*>).updateAllItems(items)
        }

        reloadBadges()
        mayBeEmpty()
    }

    private fun reloadBadges() {
        if (appSettingsService.isDisplayUnreadCountEnabled() || appSettingsService.isDisplayAllCountEnabled()) {
            CoroutineScope(Dispatchers.Main).launch {
                repository.reloadBadges()
                reloadBadgeContent()
            }
        }
    }

    private fun reloadBadgeContent() {
        if (appSettingsService.isDisplayUnreadCountEnabled()) {
            tabNewBadge
                .setText(repository.badgeUnread.toString())
                .maybeShow()
        }
        if (appSettingsService.isDisplayAllCountEnabled()) {
            tabArchiveBadge
                .setText(repository.badgeAll.toString())
                .maybeShow()
            tabStarredBadge
                .setText(repository.badgeStarred.toString())
                .maybeShow()
        }
    }

    private fun reloadTagsBadges() {
        tagsBadge.forEach {
            binding.mainDrawer.updateBadge(it.key, StringHolder(it.value.toString()))
        }
        binding.mainDrawer.resetDrawerContent()
    }

    private fun calculateNoOfColumns(): Int {
        val displayMetrics = resources.displayMetrics
        val dpWidth = displayMetrics.widthPixels / displayMetrics.density
        return (dpWidth / 300).toInt()
    }

    override fun onQueryTextChange(p0: String?): Boolean {
        if (p0.isNullOrBlank()) {
            repository.searchFilter = null
            getElementsAccordingToTab()
            fetchOnEmptyList()
        }
        return false
    }

    override fun onQueryTextSubmit(p0: String?): Boolean {
        repository.searchFilter = p0
        getElementsAccordingToTab()
        fetchOnEmptyList()
        return false
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        val inflater = menuInflater
        inflater.inflate(R.menu.home_menu, menu)

        val searchItem = menu.findItem(R.id.action_search)
        val searchView = searchItem.getActionView() as SearchView
        searchView.setOnQueryTextListener(this)

        return true
    }

    private fun needsConfirmation(titleRes: Int, messageRes: Int, doFn: () -> Unit) {
        AlertDialog.Builder(this@HomeActivity)
            .setMessage(messageRes)
            .setTitle(titleRes)
            .setPositiveButton(android.R.string.ok) { _, _ -> doFn() }
            .setNegativeButton(android.R.string.cancel) { _, _ -> }
            .create()
            .show()
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        when (item.itemId) {
            R.id.refresh -> {
                needsConfirmation(R.string.menu_home_refresh, R.string.refresh_dialog_message) {
                    Toast.makeText(this, R.string.refresh_in_progress, Toast.LENGTH_SHORT).show()
                    // TODO: Use Dispatchers.IO
                    CoroutineScope(Dispatchers.Main).launch {
                        val updatedRemote = repository.updateRemote()
                        if (updatedRemote) {
                            // TODO: Send toast messages from the repository
                            Toast.makeText(
                                this@HomeActivity,
                                R.string.refresh_success_response, Toast.LENGTH_LONG
                            )
                                .show()
                        } else {
                            Toast.makeText(
                                this@HomeActivity,
                                R.string.refresh_failer_message,
                                Toast.LENGTH_SHORT
                            ).show()
                        }
                    }
                }
                return true
            }
            R.id.readAll -> {
                if (elementsShown == ItemType.UNREAD) {
                    needsConfirmation(R.string.readAll, R.string.markall_dialog_message) {
                        binding.swipeRefreshLayout.isRefreshing = true

                        CoroutineScope(Dispatchers.Main).launch {
                            val success = repository.markAllAsRead(items)
                            if (success) {
                                Toast.makeText(
                                    this@HomeActivity,
                                    R.string.all_posts_read,
                                    Toast.LENGTH_SHORT
                                ).show()
                                tabNewBadge.removeBadge()

                                handleDrawerItems()

                                getElementsAccordingToTab()
                            } else {
                                Toast.makeText(
                                    this@HomeActivity,
                                    R.string.all_posts_not_read,
                                    Toast.LENGTH_SHORT
                                ).show()
                            }
                            handleListResult()
                            binding.swipeRefreshLayout.isRefreshing = false
                        }
                    }
                }
                return true
            }
            R.id.action_disconnect -> {
                appSettingsService.clearAll()
                val intent = Intent(this, LoginActivity::class.java)
                this.startActivity(intent)
                this@HomeActivity.finish()
                return true
            }
            else -> return super.onOptionsItemSelected(item)
        }
    }

    private fun maxItemNumber(): Int =
        when (elementsShown) {
            ItemType.UNREAD -> repository.badgeUnread
            ItemType.ALL -> repository.badgeAll
            ItemType.STARRED -> repository.badgeStarred
            else -> repository.badgeUnread // if !elementsShown then unread are fetched.
        }

    private fun updateItems(adapterItems: ArrayList<SelfossModel.Item>) {
        items = adapterItems
    }

    private fun handleRecurringTask() {
        if (appSettingsService.isPeriodicRefreshEnabled()) {
            val myConstraints = Constraints.Builder()
                .setRequiresBatteryNotLow(true)
                .setRequiresCharging(appSettingsService.isRefreshWhenChargingOnlyEnabled())
                .setRequiresStorageNotLow(true)
                .build()

            val backgroundWork =
                PeriodicWorkRequestBuilder<LoadingWorker>(appSettingsService.getRefreshMinutes(), TimeUnit.MINUTES)
                    .setConstraints(myConstraints)
                    .addTag("selfoss-loading")
                    .build()

            WorkManager.getInstance(baseContext).enqueueUniquePeriodicWork("selfoss-loading", ExistingPeriodicWorkPolicy.KEEP, backgroundWork)
        }
    }
}

